#!/usr/bin/env python

# Everything in this file is from my memory, aside from the
# info about Python's built-in sorted() which I researched after
# writing everything else here.


# Worst case, everything is on the same side of the pivot every time.
# Then we compare this many times: (n-1) + (n-2) + (n-3) + ... + 0
# which has n terms and is O(n^2).
#
# Best case, we split the list in half each time.  Then we compare
# this many times: (n-1) + ((n-1)/2 - 1)*2 + ... =
#                  (n-1) + (n-3) + ((n-3)/4 - 1)*4 + ... =
#                  (n-1) + (n-3) + (n-7) + ... =
#                  (n - (2^1 - 1)) + (n - (2^2 - 1)) + (n - (2^3 - 1)) ...
# This series has O(log n) terms and is O(n log n).
#
# Average case, if we assume* the pivot is equally likely to be anywhere
# in the ordering, we split the list in a 1/4--3/4 or more-balanced split
# half of the time.  Such a split reduces log_2 n by at least 1/2 (compare
# to the best case 1+delta and worst case 0+delta, with delta the effect
# of removing one pivot element, which is small for large n).  So on average
# it is at least half the time at least half as good as the best case:
# that's only a constant factor of 4 worse than the best case, so it's
# O(n log n) average case.
# (The constant factor is smaller than 4 because many of the better splits
# are better than 1/4--3/4 and many of the worse splits are better than 0--1,
# but I didn't do the math to find out exactly what the factor is.)
#
# * or if we were to select which item is the pivot using a true random
# number, which would make our assumption guaranteed true (but add a large
# enough constant factor that we'd rather use another sort algorithm
# that is worst-case O(n log n) anyway).


# Data collected:
# % python --version
# Python 3.2.3
# (the program also works in python 2 unmodified).
# time python 1i.py
# python 1i.py  0.07s user 0.01s system 96% cpu 0.080 total
# I'll look at the user time.  So there's a 0.07 s constant overhead.
# I'll subtract that from each of these measured times before noting it.
#
# `time python 1i.py 10000`, for example.  I'm running each three times
# and picking the middle time value (not the most perfect but not terrible
# for an exercise).
#  10000: 0.17 s
#  20000: 0.37 s
#  40000: 0.76 s
#  80000: 1.73 s
# 160000: 3.51 s
# 320000: 7.38 s
#
# With each doubling of input size (with randomly ordered inputs),
# the runtime slightly more than doubles.  That's consistent with n log n.
# (Cache sizes and other things can screw up apparent trends, but that
# seems not to have happened too dramatically in this sample, for better or
# for worse.)
#
# Let's try the worst-case!
# `time python 1i.py 200 bad`, for example.  I'll subtract the same 0.07s
# from each.
# 200: 0.02 s
# 400: 0.06 s
# 800: 0.22 s
# As expected, each doubling is roughly multiplying the runtime by four
# (quadratic increase).  Python crashes for higher numbers than 1000 because
# of recursion depth (I think this is easy to adjust, but these numbers are
# decent enough for exposition).


def non_inplace_quicksort(items):
	"""
	This sorts lists.  Use sorted(list) instead if you
	are actually programming Python, because it has the exact
	same effect and is part of the language.  Like this
	sort implementation, it is a stable sort (as of Python 2.2
	http://wiki.python.org/moin/HowTo/Sorting#Sort_Stability_and_Complex_Sorts
	).  Unlike this sort implementation, it is faster because
	it is written in C and had a lot of work put into its performance (
	http://weblog.hotales.org/cgi-bin/weblog/nb.cgi/view/python/2002/08/03/0
	https://en.wikipedia.org/wiki/Timsort
	).

	This version is not in-place, but it is not too much harder
	to write an in-place version (one which does not allocate
	memory, aside from the stack).  It would involve swapping
	the order of elements rather than creating new lists
	to put them in.
	
	>>> non_inplace_quicksort([1, 2, 3])
	[1, 2, 3]
	>>> non_inplace_quicksort([2, 3, 2, 1])
	[1, 2, 2, 3]
	>>> non_inplace_quicksort([2, 3, 2, 1, 7, -3242])
	[-3242, 1, 2, 2, 3, 7]
	>>> non_inplace_quicksort([])
	[]
	>>> non_inplace_quicksort(['God'])
	['God']
	"""
	# 'if' conditions and trivial returns are all O(1).
	if len(items) == 0:
		return []
	elif len(items) == 1:
		# lists are mutable in Python so we create a new list
		# to return, rather than return 'items'.
		return [items[0]]
	else:
		# choose an arbitrary element as the pivot, O(1).
		pivot = items[0]

		# Filtering/partitioning is O(n).
		# This could be more efficient by a constant factor by
		# iterating items once, not multiple times.
		prevs = list(filter(lambda x: x < pivot, items))
		nexts = list(filter(lambda x: x > pivot, items))
		# It's not necessary to separate the equal ones into their
		# own category, but it makes this a stable sort
		# (elements that are equivalent by the equivalence relation
		# are not changed in their order in the list).
		# Other things being equal, stable sorts are nice,
		# (occasionally code requires a stable sort or produces
		# prettier results with one).  Sort functions are sometimes
		# unstable in order to be faster by a constant factor.
		# (Writing 'sort' in python is probably the worst constant
		# factor here, though, assuming we're using the CPython
		# interpreter!)
		eqs = list(filter(lambda x: x == pivot, items))

		# prevs and nexts each have at most one less item than items,
		# so this algorithm is guaranteed to terminate.
		# We don't need to write 'pivot' here explicitly because
		# it will be a part of 'eqs'.
		# list append is O(n).
		return non_inplace_quicksort(prevs) \
			+ eqs \
			+ non_inplace_quicksort(nexts)

if __name__ == "__main__":
	import doctest
	doctest.testmod()

	print("(Usage: python 1i.py [test-case-size-for-timing [bad]]\n\
  'bad' means sorted data, which is worst for quicksort.)\n")
	import sys
	if len(sys.argv) > 1:
		testSize = int(sys.argv[1])
		bad = False
		testList = list(range(testSize))
		if not (len(sys.argv) > 2 and sys.argv[2] == "bad"):
			import random
			random.shuffle(testList)
		result = non_inplace_quicksort(testList)
		if testSize < 150:
			print(testList)
			print(result)

