{-

I invented this algorithm using my head.

O(n * v) where n is the the number of points provided and v is
the number of vertices in the convex polygon.

This could be made average-case O(n log n) (n = provided points)
by using a quadtree instead of linear search in nextPoint.

To test and run:
* have a GHC/Haskell-platform environment
* % cabal install blank-canvas
* % cabal install doctest
* if necessary, % cabal install random
* % ~/.cabal/bin/doctest 1iii.hs
* % runghc 1iii.hs
* after it starts running,
  it will print a debug message saying the polygon vertices.
  Then open a browser and go to localhost:3000 to look at the polygon.
Because I was lazy, it can only be tweaked by the CONFIGURATION PARAMETERS
below in this file.  Also, it generates points randomly distributed in a
rectangle, which isn't optimal for being able to tweak the number of
vertices of the resulting convex polygon (distribution within a circular
disk, for example, would be better for that particular purpose).  It was
amusing to see the runtime increasing quadratically for a little while as
I'd increased the number of points, and then decreasing because I'd had
testWidth small compared to the number of points, and when you fill up a
square area the polygon is just a quadrilateral!

I decided to not risk anything to inexactness and used Haskell's
arbitrary-precision rational arithmetic (which is implemented with
the GMP library) and avoided trigonometry functions (which produce
nonrational results).  I've done spatial algorithms like this before
and I know it's quite practical to do with fixed-size ints and doing
the algebra to get rid of actual rational arithmetic, I just didn't
do that here - it shouldn't affect the asymptotic performance since
the rationals are not growing to unbounded height (rational height
refers to max(log(numerator), log(denominator))).

Basically it starts with the bottom left point, which must be on the
polygon, then goes in a circle starting from the angle of the previous
edge (or off to the middle of nowhere to start off), and finds the
point (by brute force) that has the least angle-rotation from the
current point and previous edge angle, and draws (concludes) an edge
between those, then moves on to that point, till it gets back to the
point where it started.

-}

module Main(main) where

import Graphics.Blank
import Control.Monad
import Data.List
import System.Random
import qualified Data.Set as Set

import Debug.Trace

-- CONFIGURATION PARAMETERS
staticTest = False
smallTest = False
testWidth = if smallTest then 99 else 9999
extraPoints = if smallTest then 0 else 1000


testPoints :: [Point]
testPoints = [(3,6), (7,7), (4,4), (2,8), (6,6), (8,2), (3,4), (2.5, 6)]

-- TODO we could avoid duplicating points (duplicates happen not
-- to hurt though).
generateTestPoints :: (RandomGen g) => Int -> g -> ([Point], g)
generateTestPoints numPoints gen = let
	(gen1, gen2) = split gen
	coords = map toRational (randomRs (1,testWidth :: Integer) gen1)
	(coords1, coords2) = splitAt numPoints coords
	points = zip coords1 coords2
	in (points, gen2)

getPoints :: IO ()
getPoints
	| staticTest = return testPoints
	| otherwise = do
		numPoints <- randomRIO (3,30)
		ourTestPoints <- getStdRandom (generateTestPoints (numPoints+extraPoints))
		return ourTestPoints

main :: IO ()
main = do
	ourTestPoints <- getPoints
	let poly = findConvexPolygon ourTestPoints
	print poly
	blankCanvas 3000 $ \context -> do
		print "hi"
		send context $ do
			drawPolygonFrom ourTestPoints poly
{-
		moveTo (50,50)
		lineTo (200,100)
		lineWidth 10
		strokeStyle "red"
		stroke ()-}



drawPolygonFrom :: [Point] -> [Point] -> Canvas ()
drawPolygonFrom points poly = do
	let polySet = Set.fromList poly
	let edges = zip poly (tail (cycle poly))
	let minx = minimum (map fst poly)
	let miny = minimum (map snd poly)
	let maxx = maximum (map fst poly)
	let maxy = maximum (map snd poly)
	-- Negate y because Canvas thinks that +y is downward and math thinks
	-- +y is upward.  Displaying in the math way makes things clearer.
	let pointToDisplay (x,y) =
		(realToFrac ((x - minx) / (maxx-minx) * 200),
		 realToFrac ((miny - y) / (miny-maxy) * 200))
	forM_ edges $ \ (a, b) -> do
		beginPath ()
		moveTo (pointToDisplay a)
		lineTo (pointToDisplay b)
		lineWidth 5
		strokeStyle "blue"
		stroke ()
	forM_ points $ \p -> do
		let (x,y) = pointToDisplay p
		let ptsize = if Set.member p polySet then 6 else 3
		beginPath ()
		moveTo (x, y + ptsize)
		lineTo (x + ptsize, y)
		lineTo (x, y - ptsize)
		lineTo (x - ptsize, y)
		fillStyle "red"
		fill ()
		

-- Haskell's Rational does arbitrary-precision rational arithmetic
type Coord = Rational
type Point = (Coord, Coord)
type Offset = (Coord, Coord)
--data Slope = Slope
--	Integer --numerator
--	Integer --denominator (zero for vertical lines)
--data DirectedPoint = DirectedPoint Point Offset 
--data QuadTreeOfPoints = None | Point Point | Quad Q Q Q Q

-- lexicographic, lowest first
--sortedPoints = sort points
--pointSet = Set.fromList points
--firstPoint = head sortedPoints


{-
--we arbitrarily always go counterclockwise on the
--mathematical coordinate plane
--So we need a way to compare two offsets given a third base offset angle
--(farther winning ties in this case, for whichever meaning is winning)
nextPoint :: DirectedPoint -> Point
-- the next point's direction will be (dest - src)
nextPoint = 
-}

-- avoid trigonometry so that we never give the wrong answer from
-- rounding error
data Angle
	= XPlus Rational  --slope
	| YPlus
	| XMinus Rational --slope
	| YMinus
	deriving (Eq, Ord, Show)
type Ray = (Point, Angle)

-- | converts an offset to the angle (represented using line slopes)
-- that it can represent.
--
-- >>> offsetAngle (2,3)
-- XPlus (3 % 2)
-- >>> offsetAngle (2, -3)
-- XPlus ((-3) % 2)
-- >>> offsetAngle (-2, -3)
-- XMinus (3 % 2)
-- >>> offsetAngle (-2, 3)
-- XMinus ((-3) % 2)
-- >>> offsetAngle (0, -3)
-- YMinus
-- >>> offsetAngle (0, 3)
-- YPlus
-- >>> offsetAngle (0, 0)
-- *** Exception: offset is not an angle
offsetAngle :: Offset -> Angle
offsetAngle (x, y) = case compare x 0 of
	LT -> XMinus (y / x)
	GT -> XPlus (y / x)
	EQ -> case compare y 0 of
		LT -> YMinus
		GT -> YPlus
		EQ -> error "offset is not an angle"

pointOffset :: Point -> Point -> Offset
pointOffset (srcx, srcy) (dstx, dsty) = (dstx - srcx, dsty - srcy)

-- | cyclicLessThan changes the least (e.g. negative-infinity) point
-- of an ordered type.  For example, 'cyclicLessThan 0' makes negative
-- numbers greater than positive numbers and otherwise preserves
-- the ordering.
--
-- https://en.wikipedia.org/wiki/Cyclic_order (exam note: I invented
-- this function without the Internet and then decided it would be better
-- documented if I found out something mathematics would call it).
--
-- >>> cyclicLessThan 1 2 3
-- True
-- >>> cyclicLessThan 1 3 0
-- True
-- >>> cyclicLessThan 1 0 3
-- False
cyclicLessThan :: (Ord a) => a -> a -> a -> Bool
cyclicLessThan lowest a b = if (a < lowest) == (b < lowest)
	then a < b
	else (b < lowest)

-- | squared to avoid rounding error from sqrt
--
-- >>> offsetDistanceSquared (2, -3)
-- 13 % 1
offsetDistanceSquared :: Offset -> Coord
offsetDistanceSquared (x, y) = x*x + y*y

pointDistanceSquared :: Point -> Point -> Coord
pointDistanceSquared a b = offsetDistanceSquared (pointOffset a b)

-- |
--
-- >>> comparePointsInCounterclockwisenessFromRay ((1,1), XPlus (3/4)) (1,2) (2,100)
-- GT
comparePointsInCounterclockwisenessFromRay :: Ray -> Point -> Point -> Ordering
comparePointsInCounterclockwisenessFromRay (base, angle) point1 point2 = let
	offset1 = pointOffset base point1
	offset2 = pointOffset base point2
	angle1 = offsetAngle offset1
	angle2 = offsetAngle offset2
	in if angle1 == angle2
		then EQ
		else case cyclicLessThan angle angle1 angle2 of
			True -> LT
			False -> GT

-- | less is better
--
-- >>> betterNextPolygonPoint ((1,1), XPlus (3/4)) (1,2) (2,100)
-- GT
-- >>> betterNextPolygonPoint ((1,1), XPlus (3/4)) (1,2) (1,100)
-- GT
-- >>> betterNextPolygonPoint ((1,1), XPlus (3/4)) (1,2) (0,100)
-- LT
betterNextPolygonPoint :: Ray -> Point -> Point -> Ordering
betterNextPolygonPoint ray@(base, _angle) point1 point2 =
	case comparePointsInCounterclockwisenessFromRay ray point1 point2 of
	LT -> LT
	GT -> GT
	EQ -> -- greater distance is better here, so that our polygon doesn't
		-- have two consecutive collinear edges, so compare in the
		-- opposite order from usual.
		compare
			(pointDistanceSquared base point2)
			(pointDistanceSquared base point1)
	

-- for ties, picks the furthest away one
-- goes (counter?)clockwise.
-- obvs never picks the same point?
nextPoint :: [Point] -> Point -> Angle -> Point
nextPoint allPoints currentPoint priorDirection =
	minimumBy
		(betterNextPolygonPoint (currentPoint, priorDirection))
		(filter (/= currentPoint) allPoints)

-- Lowest x, with lowest y for ties, will always be one of the
-- polygon vertices.
--
-- O(n)
firstPoint :: [Point] -> Point
firstPoint points = minimum points

-- Starting from the furthest left (and bottom) point,
-- the exact direction doesn't matter too much;
-- it can be anything not pointing into the field of points.
firstDirection :: Angle
firstDirection = XMinus 0

-- TODO What about non-polygons? (e.g. < 3 sides)
findConvexPolygon :: [Point] -> [Point]
findConvexPolygon allPoints = let
	firstVertex = firstPoint allPoints
	go verticesSoFar currentVertex priorDirection =
		--trace (show ("yo!", currentVertex, priorDirection)) $
		let
		nextVertex = nextPoint allPoints currentVertex priorDirection
		in if nextVertex == firstVertex
		then verticesSoFar
		else go (nextVertex : verticesSoFar) nextVertex
			(offsetAngle (pointOffset currentVertex nextVertex))
	in go [firstVertex] firstVertex firstDirection


