
// Copied from 2A.js and rewritten as I go.
//
// The Javascript was somewhat more imperative-style then
// the Haskell, which means it will be somewhat easier
// to translate into C++.
// 
// % g++ -Wall -g -std=c++0x -O3 2A.cpp
// % ./a.out
//
// Tests are currently just asserts run at the beginning of main().
//
// Commented out std::cerrs were put in for debugging purposes and left
// there for anthropological observers' sake.  It turns out I was
// outputting a char as an int because I didn't cast it the right way
// and its numeric value was being output; silly weakly typed C++.
//
// Valgrind spies nothing wrong.

#include <vector>
#include <map>
#include <algorithm>
#include <iterator>
#include <iostream>
#include <cassert>

using std::vector;
using std::map;

vector<int> findPerfect3DigitSquares() {
  vector<int> result;
  for(int i = 10; i != 32; ++i) {
    result.push_back(i*i);
  }
  return result;
}
const vector<int> perfect3DigitSquares = findPerfect3DigitSquares();

typedef int digit;
vector<digit> asDigits(int integer) {
  vector<digit> result;
  if(integer == 0) {
    result.push_back(0);
  }
  else {
    while(integer != 0) {
      digit d = integer % 10;
      result.push_back(d);
      integer = (integer - d) / 10;
    }
    std::reverse(result.begin(), result.end());
  }
  return result;
}



vector<vector<digit>> findPerfect3DigitSquaresAsDigits() {
  vector<vector<digit>> result;
  for(int i : perfect3DigitSquares) {
    result.push_back(asDigits(i));
  }
  return result;
}
const vector<vector<digit>> perfect3DigitSquaresAsDigits =
                              findPerfect3DigitSquaresAsDigits();

/*


      *   * * *   * * *   *
      * * *   * * *   * * *
      *   * * *   * * *   *

grid: coordinates from top-left starting at (0,0),
number positions indicated by the top-left of them
(and implicitly whether it is across or down, and
we know it's 3 long).

      0 1 2 3 4 5 6 7 8 9 10
    0 *   * * *   * * *   *
    1 * * *   * * *   * * *
    2 *   * * *   * * *   *

*/

typedef int Coord;
struct Loc {
  Loc(Coord x, Coord y) : x(x), y(y) {}
  Coord x;
  Coord y;
};
const vector<Loc> rows =
  {Loc(2,0),Loc(6,0),Loc(0,1),Loc(4,1),Loc(8,1),Loc(2,2),Loc(6,2)};
const vector<Loc> cols =
  {Loc(0,0),Loc(2,0),Loc(4,0),Loc(6,0),Loc(8,0),Loc(10,0)};

// Use a tree map - less-than is slightly easier to implement than hash
// and it doesn't matter (and the hash-based unordered_map has only been
// standardized for a few years).
// Luckily, in C++, like Haskell and unlike Javascript, anything can be
// a key of the standard dictionary type!
// (You just have to implement the comparison functions.)
inline bool operator==(Loc a, Loc b) {
  return a.x == b.x && a.y == b.y;
}
inline bool operator<(Loc a, Loc b) {
  return a.x < b.x || (a.x == b.x && a.y < b.y);
}
typedef map<Loc, digit> Board; 

// C++ has similar lack of functional ways of doing things to JavaScript,
// and we already imperative-ized this for JS, so it's not hard.
// On the other hand, STL iterators make the code a little messier looking.
// Luckily C++11 has 'auto', a limited form of type inference, and
// range-based for loops, which make this code more readable.
//
// Since this can work on any associative container, why write this:
//bool areConsistent(Board const& board1, Board const& board2) {
// when we could write it polymorphically:
template<typename Assoc>
bool areConsistent(Assoc const& board1, Assoc const& board2) {
  for(auto pair : board1) {
    auto i = board2.find(pair.first);
    if(i != board2.end() && i->second != pair.second) {
      return false;
    }
  }
  return true;
}


// There are 22 possible 3-digit squares,
// and 13 slots to put squares in.
// If we just tried every combination, there are 22^13 of them,
// which is too many.
// Each placement, however, rules out a lot of possibilities,
// so a depth first search taking the slots in an order in which
// they're connected might be fast enough.

// A slot is three locs in a row, in an array.
typedef vector<Loc> Slot;
Slot rowToSlot(Loc loc) {
  return vector<Loc>{Loc(loc.x, loc.y), Loc(loc.x+1, loc.y), Loc(loc.x+2, loc.y)};
}
Slot colToSlot(Loc loc) {
  return vector<Loc>{Loc(loc.x, loc.y), Loc(loc.x, loc.y+1), Loc(loc.x, loc.y+2)};
}
vector<Slot> findSlots(vector<Loc> const& rows, vector<Loc> const& cols) {
  vector<Slot> result;
  for(Loc loc : rows) { result.push_back(rowToSlot(loc)); }
  for(Loc loc : cols) { result.push_back(colToSlot(loc)); }
  return result;
}
const vector<Slot> slots = findSlots(rows, cols);

// Sorting by (x+y) should be good enough for an efficient depth-first
// search ordering for this problem.

vector<Slot> sortedForSearch(vector<Slot> slots) {
  std::sort(slots.begin(), slots.end(), [](Slot const& slot1, Slot const& slot2) {
    return (slot1[0].x + slot1[0].y) < (slot2[0].x + slot2[0].y);
  });
  return slots;
}
const vector<Slot> roughlySortedSlots = sortedForSearch(slots);

// we want:
// const Board solution = ...;

void showBoard(Board const& board, std::ostream& os);
// There's no easy way for C++ to represent a lazy list, so I'll just
// write this to return the first answer.
Board solve(Board const& boardSoFar, vector<Slot> slotsStillToFill) {
  //showBoard(boardSoFar, std::cerr);
  //std::cerr << "\n===" << slotsStillToFill.size() << "\n\n";
  if(slotsStillToFill.size() == 0) {
    return boardSoFar;
  }
  vector<Board> nextBoards;
  const Slot nextSlot = slotsStillToFill[0];
  for(vector<digit> square : perfect3DigitSquaresAsDigits) {
    Board miniboard;
    assert(nextSlot.size() == square.size());
    for(size_t i = 0; i != nextSlot.size(); ++i) {
      //std::cerr << "??" << square[i] << "\n";
      miniboard[nextSlot[i]] = square[i];
    }
    if(areConsistent(boardSoFar, miniboard)) {
      nextBoards.push_back(miniboard);
      std::copy(boardSoFar.begin(), boardSoFar.end(),
                std::inserter(nextBoards.back(), nextBoards.back().end()));
    }
  }
  // Does this copying-based impl subSlots make the runtime
  // asymptotically worse?  I'm not sure.
  // I could pass around an index into the array instead of slicing it.
  const vector<Slot> subSlots(slotsStillToFill.begin()+1, slotsStillToFill.end());
  for(Board const& board : nextBoards) {
    Board solvedBoard = solve(board, subSlots);
    if(solvedBoard.size() != 0) {
      return solvedBoard;
    }
  }
  return Board();
}

const Board solution = solve(Board(), roughlySortedSlots);

void showBoard(Board const& board, std::ostream& os) {
  Coord minx = std::numeric_limits<Coord>::max();
  Coord miny = std::numeric_limits<Coord>::max();
  Coord maxx = std::numeric_limits<Coord>::min();
  Coord maxy = std::numeric_limits<Coord>::min();
  for(auto const& p : board) {
    const Loc loc = p.first;
    //std::cerr << "??" << loc.x << "?" << loc.y << "??" << p.second << "\n";
    if(minx > loc.x) { minx = loc.x; }
    if(miny > loc.y) { miny = loc.y; }
    if(maxx < loc.x) { maxx = loc.x; }
    if(maxy < loc.y) { maxy = loc.y; }
  }
  for(Coord y = miny; y <= maxy; ++y) {
    for(Coord x = minx; x <= maxx; ++x) {
      auto i = board.find(Loc(x,y));
      if(i == board.end()) { os << ' '; }
      else { os << char('0' + i->second); }
    }
    os << '\n';
  }
}

int main() {
  assert((asDigits(0) == vector<digit>{0}));
  assert((asDigits(13) == vector<digit>{1,3}));
  assert((asDigits(127) == vector<digit>{1,2,7}));

  {
  // Well, this is a hacky-looking way to give C++ dictionary literals,
  // but in C++11, it works!
  typedef std::pair<char, int> P;
  typedef std::map<char, int> M;
  assert(!areConsistent(M{P('a', 5), P('b', 7), P('c', 8)},
                        M{P('a', 5), P('b', 2), P('d', 9)}));
  assert(areConsistent(M{P('a', 5), P('b', 7), P('c', 8)},
                       M{P('a', 5), P('b', 7), P('d', 9)}));
  }

  showBoard(solution, std::cout);
}
