
--This file is a stream of consciousness:
--I wrote it from top to bottom (aside from the imports list),
--stopping many times along the way to typecheck, test functions
--in the interpreter GHCI, and think about what a nice next step is.


import Data.Char
import Data.List
import qualified Data.Map as Map
import Data.Map(Map)

import Test.QuickCheck

-- GHCI + knowing their bases
perfect3DigitSquares :: [Int]
perfect3DigitSquares = map (\n->n^2) [10..31]

-- since we don't care about efficiency here, just use strings
asDigits :: Int -> [Int]
asDigits = map digitToInt . show

-- tested in GHCI
perfect3DigitSquaresAsDigits :: [[Int]]
perfect3DigitSquaresAsDigits = map asDigits perfect3DigitSquares

{-


      *   * * *   * * *   *
      * * *   * * *   * * *
      *   * * *   * * *   *

grid: coordinates from top-left starting at (0,0),
number positions indicated by the top-left of them
(and implicitly whether it is across or down, and
we know it's 3 long).

      0 1 2 3 4 5 6 7 8 9 10
    0 *   * * *   * * *   *
    1 * * *   * * *   * * *
    2 *   * * *   * * *   *

-}

-- location : x,y
type Loc = (Int,Int)
rows :: [Loc]
rows = [(2,0),(6,0),(0,1),(4,1),(8,1),(2,2),(6,2)]
cols :: [Loc]
cols = [(0,0),(2,0),(4,0),(6,0),(8,0),(10,0)]

type Board = Map Loc Int

-- |
-- (Yes, Haskell's dictionary literal syntax sucks in its nonexistence:)
--
-- >>> areConsistent (Map.fromList [('a',5),('b',7),('c',8)]) (Map.fromList [('a',5),('b',2),('d',9)])
-- False
-- >>> areConsistent (Map.fromList [('a',5),('b',7),('c',8)]) (Map.fromList [('a',5),('b',7),('d',9)])
-- True
--
-- (Dear readers of the printed version: I couldn't find a way
--  that worked to split each test over more than one line.
--  The first one's argument lists are
--    [('a',5),('b',7),('c',8)] and [('a',5),('b',2),('d',9)]
--  and the second one's argument lists are
--    [('a',5),('b',7),('c',8)] and [('a',5),('b',7),('d',9)].
-- )
--
-- (to run these tests:
-- % cabal install doctest
-- % ~/.cabal/bin/doctest 2A.hs
-- )
--
-- areConsistent :: Board -> Board -> Bool
areConsistent :: (Ord k, Eq v) => Map k v -> Map k v -> Bool
areConsistent b1 b2 = and (Map.elems (Map.intersectionWith (==) b1 b2))


-- There are 22 possible 3-digit squares,
-- and 13 slots to put squares in.
-- If we just tried every combination, there are 22^13 of them,
-- which is too many.
-- Each placement, however, rules out a lot of possibilities,
-- so a depth first search taking the slots in an order in which
-- they're connected might be fast enough.

-- three locs in a row:
type Slot = [Loc]
rowToSlot :: Loc -> Slot
rowToSlot (x,y) = [(x,y), (x+1,y), (x+2,y)]
colToSlot :: Loc -> Slot
colToSlot (x,y) = [(x,y), (x,y+1), (x,y+2)]
slots :: [Slot]
slots = map rowToSlot rows ++ map colToSlot cols

-- Sorting by (x+y) should be good enough for an efficient depth-first
-- search ordering for this problem.
roughlySortedSlots :: [Slot]
roughlySortedSlots = sortBy
  (\[(x1,y1),_,_] [(x2,y2),_,_] -> compare (x1+y1) (x2+y2))
  slots

-- we want:
-- solution :: Board

-- We might as well solve for all the boards, because Haskell's laziness
-- means that if we only ask for the first one then we won't need to compute
-- all possible solutions.

solve :: Board -> [Slot] -> [Board]
solve boardSoFar [] = [boardSoFar]
solve boardSoFar (slot:slotsStillToFill) = let
  nextBoards = [Map.union boardSoFar miniboard
               | square <- perfect3DigitSquaresAsDigits,
                 let miniboard = Map.fromList (zip slot square),
                 areConsistent boardSoFar miniboard]
  in concat (map (\board -> solve board slotsStillToFill) nextBoards)

solution :: Board
solution = head (solve Map.empty roughlySortedSlots)
-- indeed using 'slots' makes this too slow; the sorting helped!
-- But now how will I check if it's correct?  I need a way to print it.

showBoard :: Board -> String
showBoard board = let
  minx = minimum (map fst (Map.keys board))
  miny = minimum (map snd (Map.keys board))
  maxx = maximum (map fst (Map.keys board))
  maxy = maximum (map snd (Map.keys board))
  xs = [minx..maxx]
  ys = [miny..maxy]
 in unlines (map (\y -> map (\x ->
    case Map.lookup (x,y) board of
      Nothing -> ' '
      Just n -> intToDigit n
  ) xs) ys)

{-
Now, in ghci,
> putStrLn (showBoard solution)
1 225 121 1
225 729 900
1 676 676 0

YAY IT WORKED!
On the first try after it typechecked, at that.
I could never have done it so easily in a less type-focused language.
Most of the time coding was spent thinking about what I actually wanted.
I would likely have done it in Haskell first even if I was ultimately aiming
to run it in another language, because Haskell was a language that
let me think about this problem.
-}

--might as well have a 'main' that shows the solution.
main :: IO ()
main = putStrLn (showBoard solution)

-- Now how will I write this in C/C++ and how will I write it in JavaScript?
-- JavaScript can do most of the things I do here; some of the Map operations
-- will be trickier (JS doesn't provide e.g. intersectionWith!) but I can
-- emulate this in a functional style in JS.  I'm not sure whether it will be
-- easy to document without types, though.
--
-- C/C++ is more imperative.  They lack garbage collection and lack
-- functional-style libraries.  I might have to implement 'solve' using
-- loops and recursion (using the stack as a part of the depth first
-- searching) - I guess that is not terribly different from the Haskell.

-- Haskell tests:
-- Sometimes Haskell is tested with QuickCheck, which tests
-- mathematical properties by trying them with several randomly generated
-- examples.
--
-- To run these:
-- % cabal install quickcheck
-- % cabal install quickcheck-script
-- % ~/.cabal/bin/quickCheck 2A.hs
prop_ConsistentWithSelf :: [(Int,Int)] -> Bool
prop_ConsistentWithSelf m =
  areConsistent m' m' where m' = Map.fromList m

prop_ConsistentWithEmpty :: [(Int,Int)] -> Bool
prop_ConsistentWithEmpty m =
  areConsistent m' Map.empty where m' = Map.fromList m

prop_inconsistentWithDifferentVal :: (Int,Int) -> Bool
prop_inconsistentWithDifferentVal (k,v) =
  not (areConsistent (Map.singleton k v) (Map.singleton k (v+1)))

prop_consistentWithDifferentPlace :: (Int,Int) -> Bool
prop_consistentWithDifferentPlace (k,v) =
  areConsistent (Map.singleton k v) (Map.singleton (k+1) v)

-- Extensibility:
-- Several things like, say, 'rows' would change with a different board.
-- All of those bindings would probably go inside a single function that
-- took that information, then.  'solve' relies on the three digit squares
-- but nothing else -- it could sensibly be tweaked to observe the size of
-- a slot and use corresponding-length squares.

