
// Copied from 2A.hs and rewritten as I go.
// 
// Run in e.g. node.  In this directory (npm installs
// in ./node_modules by default), do:
//
// % npm install underscore
// % node 2A.js
//
// Tests are currently just asserts run when this file is loaded.

var _ = require('underscore');
var assert = require('assert');

var perfect3DigitSquares = _.map(_.range(10,32), function(n) { return n*n; });

function asDigits(integer) {
  // Since JS's strings are nontrivial to use as arrays I'll just write
  // the better, non-string-using, code here.
  if(integer === 0) {
    return [0];
  }
  var result = [];
  while(integer !== 0) {
    var digit = integer % 10;
    result.push(digit);
    integer = (integer - digit) / 10;
  }
  return result.reverse();
}

assert.deepEqual(asDigits(0), [0]);
assert.deepEqual(asDigits(13), [1,3]);
assert.deepEqual(asDigits(127), [1,2,7]);


var perfect3DigitSquaresAsDigits = _.map(perfect3DigitSquares, asDigits);

/*


      *   * * *   * * *   *
      * * *   * * *   * * *
      *   * * *   * * *   *

grid: coordinates from top-left starting at (0,0),
number positions indicated by the top-left of them
(and implicitly whether it is across or down, and
we know it's 3 long).

      0 1 2 3 4 5 6 7 8 9 10
    0 *   * * *   * * *   *
    1 * * *   * * *   * * *
    2 *   * * *   * * *   *

*/

// location : x,y
// [x,y] or {x:x, y:y} would be reasonable
// function loc(x, y) { return [x, y]; }
function mkloc(x, y) { return { x: x, y: y }; }
var rows =
  [mkloc(2,0),mkloc(6,0),mkloc(0,1),mkloc(4,1),mkloc(8,1),mkloc(2,2),mkloc(6,2)];
var cols =
  [mkloc(0,0),mkloc(2,0),mkloc(4,0),mkloc(6,0),mkloc(8,0),mkloc(10,0)];

// The board is a hash from loc to digit... but wait, JS objects can only
// have strings, not pairs, as keys!  So we'll have a way to transform
// locs to hash key strings and back.
//
// (I might not have realized this up-front if I hadn't been looking at
// the data declaration in Haskell.  Instead I would've realized it while
// writing the next function or so, and had to go back and write this then.)
function locKey(loc) { return loc.x + "," + loc.y; }
//(debugging: this needs to turn the x/y strings into numbers,
//and using 'map' makes this long enough that the functional style
//is not worth it anymore.)
//function keyLoc(key) { return mkloc.apply(null, key.split(',')); }
function keyLoc(key) {
  var xy = key.split(',');
  return mkloc(+xy[0], +xy[1]);
}

// JS doesn't come with anything like intersectionWith.
function areConsistent(board1, board2) {
  for(var l in board1) if(_.has(board1, l)) {
    if(_.has(board2, l) && board2[l] !== board1[l]) {
      return false;
    }
  }
  return true;
}

// But at least JS has dictionary literals.
assert(!areConsistent({'a': 5, 'b': 7, 'c': 8}, {'a': 5, 'b': 2, 'd': 9}),
         "inconsistent boards");
assert(areConsistent({'a': 5, 'b': 7, 'c': 8}, {'a': 5, 'b': 7, 'd': 9}),
         "consistent boards");

// There are 22 possible 3-digit squares,
// and 13 slots to put squares in.
// If we just tried every combination, there are 22^13 of them,
// which is too many.
// Each placement, however, rules out a lot of possibilities,
// so a depth first search taking the slots in an order in which
// they're connected might be fast enough.

// A slot is three locs in a row, in an array.
function rowToSlot(loc) {
  return [mkloc(loc.x, loc.y), mkloc(loc.x+1, loc.y), mkloc(loc.x+2, loc.y)];
}
function colToSlot(loc) {
  return [mkloc(loc.x, loc.y), mkloc(loc.x, loc.y+1), mkloc(loc.x, loc.y+2)];
}
var slots = _.map(rows, rowToSlot).concat(_.map(cols, colToSlot));

// Sorting by (x+y) should be good enough for an efficient depth-first
// search ordering for this problem.
var roughlySortedSlots = slots.sort(function(slot1, slot2) {
    return (slot1[0].x + slot1[0].y) - (slot2[0].x + slot2[0].y);
  });

// we want:
// var solution = a board

// There's no easy way for JS to represent a lazy list, so I'll just
// write this to return the first answer.
function solve(boardSoFar, slotsStillToFill) {
  //(debugging... this shows it's resulting in the
  //right thing, so the bug must be in showBoard:
  //console.log("step", boardSoFar, slotsStillToFill);)
  if(slotsStillToFill.length === 0) {
    return boardSoFar;
  }
  var nextBoards = [];
  _.each(perfect3DigitSquaresAsDigits, function(square) {
    var miniboard = _.object(_.map(slotsStillToFill[0], locKey), square);
    if(areConsistent(boardSoFar, miniboard)) {
      nextBoards.push(_.extend({}, boardSoFar, miniboard));
    }
  });
  for(var i = 0; i !== nextBoards.length; ++i) {
    // Does this copying-based impl of slotsStillToFill recursion
    // make the runtime asymptotically worse?  I'm not sure.
    // I could pass around an index into the array instead of slicing
    // it, given that JS doesn't come with linked lists.
    var solvedBoard = solve(nextBoards[i], slotsStillToFill.slice(1));
    if(solvedBoard !== null) {
      return solvedBoard;
    }
  }
  return null;
}

var solution = solve({}, roughlySortedSlots);

function getx(loc) {
  return loc.x;
}
function gety(loc) {
  return loc.y;
}

function showBoard(board) {
  var locs = _.map(_.keys(board), keyLoc);
  var minx = _.min(_.map(locs, getx));
  var miny = _.min(_.map(locs, gety));
  var maxx = _.max(_.map(locs, getx));
  var maxy = _.max(_.map(locs, gety));
  // debugging:
  // I introduced this to debug another thing and now
  // it's informing me that maxx is 9 when it should be
  // 10 (though the rest are correct and solution is
  // correct).  Perhaps the two digits are causing problems
  // somehow?  Ah - whoops we're doing min/max on strings!
  // (bug in keyLoc due to lack of type safety.)
  // Yay that fixed the program and it outputs the right
  // answer now.
  console.log(minx, miny, "&&", maxx, maxy);
  var xs = _.range(minx, maxx+1);
  var ys = _.range(miny, maxy+1);
  var lines = _.map(ys, function(y) {
    return "".concat.apply("", _.map(xs, function(x) {
      var tile = board[locKey(mkloc(x,y))];
      if(tile === undefined) { return ' '; }
      else { return ""+tile; }
    })) + '\n';
  });
  return "".concat.apply("", lines);
}

//might as well have a 'main' that shows the solution.
function main() {
  console.log(showBoard(solution));
}

main();

// Blurgh, even after fixing JsHint warnings and runtime
// exceptions, it's not outputing the right value.

// Fixes are described in comments that mention the word 'debugging'.
// It works now.

