
// I might as well do this in C to see how that is.

// Usage:
//   gcc -Wall -std=c99 -g 2B.c
//   cat 2B.csv | ./a.out | dot -Tpng | display -

// Testing:
//   cat 2B.csv | valgrind ./a.out
// shows that it does not read any uninitialized memory
// or write any unallocated memory.
// It outputs the right thing.
// It could use better tests, but then again, it could use
// all the known bugs in it fixed too.
//
// C (and C++) have great tooling (gdb! valgrind! kcachegrind!
// lots of IDEs!) but they don't have as easy and fluid use of
// libraries as javascript (NodeJS, npm, or copying/linking js
// libs in the browser) or Haskell (cabal) or several other of
// the scripting languages.  This might be hip people's fault
// for not letting C be hip; someone attempted to have a CPAN-for-C
// named CCAN: http://ccodearchive.net/ which is small and I haven't
// tried but I wish that became easy and hip.  Or maybe C just
// isn't as good a language for writing libraries?

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <malloc.h>

// C does not have hashes or binary search trees built in.
// I'll just use linear-time searches for this example,
// though a good C program should use a library rather than
// have the overall program be n^2 complexity.
// ... actually, I didn't even do any linear searches, instead
// I just made the output slightly different.

// It also does not free memory, because the program terminates
// promptly anyway.  The next time I write a program like this,
// I should use Go, because it is at a similar abstraction level but
// uses garbage collection.

size_t countCharOccurrence(const char * text, char counted) {
	size_t count = 0;
	while(*text != '\0') {
		count += (*text++ == counted);
	}
	return count;
}
size_t countCharOccurrenceRange(const char * text, const char * end, char counted) {
	size_t count = 0;
	while(text != end) {
		count += (*text++ == counted);
	}
	return count;
}

struct row {
	const char * id;
	const char * name;
	const char ** parents; // null-terminated
	const char * birthdate;
	const char * deathdate;
};

const char * alloc_str(const char * begin, size_t len) {
	char * str = malloc(len + 1);
	memcpy(str, begin, len);
	str[len] = '\0';
	return str;
}
const char * alloc_str_range(const char * begin, const char * end) {
	assert(begin <= end);
	return alloc_str(begin, end - begin);
}

struct row parseRow(const char * begin, size_t len) {
	// The error handling here is pretty bad.
	const char * id_end = strchr(begin, ';');
	const char * name_end = strchr(id_end+1, ';');
	const char * parents_end = strchr(name_end+1, ';');
	const char * birthdate_end = strchr(parents_end+1, ';');
	const char * deathdate_end = begin + len;
	assert(deathdate_end > birthdate_end && "Not enough fields in row");
	struct row result;
	result.id = alloc_str_range(begin, id_end);
	result.name = alloc_str_range(id_end+1, name_end);
	result.birthdate = alloc_str_range(parents_end+1, birthdate_end);
	result.deathdate = alloc_str_range(birthdate_end+1, deathdate_end);
	
	size_t num_parents = 1 + countCharOccurrenceRange(name_end+1, parents_end, ',');
	result.parents = malloc((num_parents+1)*sizeof(const char *));
	
	const char * parent_begin = name_end+1;
	size_t parent_num = 0;
	if(parent_begin != parents_end) {
		for(const char * i = parent_begin; i != parents_end; ++i) {
			if(*i == ',') {
				result.parents[parent_num++] =
					alloc_str_range(parent_begin, i);
				parent_begin = i+1;
			}
		}
		result.parents[parent_num++] = alloc_str_range(parent_begin, parents_end);
	}
	result.parents[parent_num] = NULL;
	return result;
}

// returns a malloc-allocated array of rows
// and puts the count of rows in 'len'
struct row* parseCsvText(const char * text, size_t* len) {
	const size_t num_lines = countCharOccurrence(text, '\n');
	// possibly an overestimate, if there are blank lines:
	struct row* rows = malloc((num_lines+1)*sizeof(struct row));
	size_t row_num = 0;

	size_t line_begin = 0;
	for(size_t i = 0; text[i] != '\0'; ++i) {
		// If the file had \r\n endings this would make the last
		// field have a \r (BUG!)
		if(text[i] == '\n') {
			if(i == line_begin) {
				continue;
			}
			rows[row_num++] = parseRow(text + line_begin, i - line_begin);
			line_begin = i + 1;
		}
	}
	*len = row_num;
	return rows;
}

// I'm not bothering to realloc or anything.
#define ARBITRARY_FILE_LENGTH_LIMIT 100000

int main() {
	char data[ARBITRARY_FILE_LENGTH_LIMIT+1];
	size_t text_len = fread(data, 1, ARBITRARY_FILE_LENGTH_LIMIT, stdin);
	data[text_len] = '\0';
	// Even in my laziness I don't like silent incorrect behavior (truncating).
	assert(text_len < ARBITRARY_FILE_LENGTH_LIMIT && "file too long for us!");
	size_t rows_len;
	struct row* rows = parseCsvText(data, &rows_len);
	puts("digraph \"family tree\" {\n");
	for(size_t i = 0; i != rows_len; ++i) {
		printf("\"%s\" [label=\"%s\"]\n", rows[i].id, rows[i].name);
		for(size_t j = 0; rows[i].parents[j]; ++j) {
			printf("\"%s\" -> \"%s\"\n", rows[i].parents[j], rows[i].id);
		}
	}
	puts("}\n");
	return 0;
}

